#!/usr/bin/perl -w
#
# taxonoTree.pl v1.0
# Last modified April 19, 2006 12:15
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose: Constructs the tree associated with a hierarchical taxonomy
#
# Input: A tab-delimited file containing species names (in first column) and the higher-level taxa
#		 they belong to (remaining columns, in order of increasing inclusiveness)
#
# Output: A nexus-formatted treefile (optionally with node labels) with the tree associated with
#		  the input taxonomy
#
# Usage: taxonoTree.pl -t<filename> [-l] [-v] [-h]
#	options: -t<filename> = tab-delimited file containing species names (in first column) and the higher-level
#							taxa they belong to (remaining columns, in order of increasing inclusiveness)
#			 -l = include names of higher level taxa as node labels
#            -v = verbose output
#            -h = print this message and quit

use strict;
use File::Basename;

# Set user-input defaults and associated parameters
	# I/O variables
		my $taxFile;
		my ($treeFile, $treeName);
		my $nodeLabels = 0;
		
	# taxonoTree variables
		my $maxLevel = 0;
		my (%taxLevel, %levelNames, %daughterTaxa);

	# Miscellaneous variables
		my $verbose = 0;
		my $debug = 0;
		my $version = "1.0";

# Process flags
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] =~ /^-t(.+)/)
			{
			$taxFile = $1;
			die "ERROR: Text file $taxFile does not exist.\n" unless (-e $taxFile);
			my ($dataSource, $dir, $ext) = fileparse($taxFile, qr/\.\D.*/);
				$treeFile = $dataSource . "_tt.tre";
				$treeName = $dataSource;
			}
		elsif ($ARGV[$i] eq "-l")
			{
			$nodeLabels = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: taxonoTree.pl -t<filename> [-l] [-v] [-h]\n";
			print "Version: $version\n";
			print "Options: -t<filename> = tab-delimited file containing species names (in first column) and the higher-level\n";
			print "                        taxa they belong to (remaining columns, in order of increasing inclusiveness)\n";
			print "         -l = include names of higher level taxa as node labels\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: taxonoTree.pl -t<filename> [-l] [-v] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}

	die "ERROR: Must specify file containing user taxonomy.\n" if (not defined $taxFile);

# Read in user taxonomy and process
	setLineBreak($taxFile);
		open (TAX, "<$taxFile") or die "Cannot open file containing user taxonomy, $taxFile\n";
			while (<TAX>)
				{
				chomp;
				next unless ($_);
				next if ($_ =~ /species/i);
				
				my @taxLine = split(/\t/, $_);
				my $lastName = "";
				
				for (my $i = 0; $i < scalar(@taxLine); $i++)
					{
					my $readName = $taxLine[$i];
						next unless ($readName);
						$readName .= "_tt$i" unless ($i == 0);	# Protection for names occurring at multiple levels
					
					
					$maxLevel = $i if ($i > $maxLevel);
					
					if (not defined $taxLevel{$readName})
						{
						$taxLevel{$readName} =  $i;
						push @ { $levelNames{$i} }, $readName unless ($i == 0);
						}
						
					push @ { $daughterTaxa{$readName} }, $lastName if ($lastName);	# Store last non-blank name as daughter taxon (account for blank entries)
						if ($debug)
							{
							print "\t" if ($i > 0);
							print "Name: $readName (level $i)";
							print "; immediate daughter: $lastName" if ($lastName);
							print "\n";
							}
						$lastName = $readName;
					}
				}
		close TAX;
	
# Build tree and substitute in taxa hierarchically
	# Build most inclusive level
		my $tree = "(" . join(',', sort(@ { $levelNames{$maxLevel} })) . ")";
			$tree .= "'root'" if ($nodeLabels);
			
	# Substitute taxa in descending order of hierarchy
		for (my $level = $maxLevel; $level >= 1; $level--)
			{
			foreach my $name (@ { $levelNames{$level} })
				{
				my @daughters = duplStrip(@ { $daughterTaxa{$name} });
				my $daughter = join(',', sort(@daughters));
					if (scalar(@ { $daughterTaxa{$name} }) > 1)
						{
						$daughter = "(" . $daughter . ")";
						if ($nodeLabels)
							{
							(my $label = $name) =~ s/_tt\d+$//;
							$daughter .= "'$label'";
							}
						}
						
				print "Name: $name; daughter: $daughter\n" if ($debug);
				
				# Perform substitution
					$tree =~ s/$name,/$daughter,/;
					$tree =~ s/$name\)/$daughter\)/;
				}
			}

# Print tree!
	open (TREE, ">$treeFile") or die "Cannot write tree to file $treeFile.\n";
		print TREE "#nexus\n\nbegin trees;\n";
		print TREE "\ttree $treeName = [&R] $tree;\n";
		print TREE "end;\n";
	close TREE;

# Subroutines used in the script

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		#Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
# 	v1.0 (April 19, 2006)
#		- initial release (based on an idea / request from Blaise Alako)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
